@@ -8,18 +8,17 @@ from inspect import FrameInfo, stack
 from enum import Enum
 from logging import Logger
 from pathlib import Path
-from pypomes_core import dict_get_key, dict_stringify, exc_format
+from pypomes_core import dict_clone, dict_stringify, exc_format
 from pypomes_db import (
     DbEngine, db_exists, db_count,
     db_select, db_insert, db_update, db_delete
 )
 from types import ModuleType
-from typing import Any, TypeVar
+from typing import Any, Literal, TypeVar
 
 from .sob_config import (
     SOB_BASE_FOLDER, SOB_MAX_THREADS,
-    sob_db_specs, sob_col_names,
-    sob_attrs_map, sob_attrs_unique
+    sob_db_columns, sob_db_specs, sob_attrs_input, sob_attrs_unique
 )
 
 # 'Sob' stands for all subclasses of 'PySob'
@@ -33,11 +32,12 @@ class PySob:
     """
 
     def __init__(self,
-                 __references: list[type] = None,
+                 __references: type[Sob | list[Sob]] | list[type[Sob | list[Sob]]] = None,
                  /,
                  where_data: dict[str, Any] = None,
                  db_engine: DbEngine = None,
                  db_conn: Any = None,
+                 committable: bool = None,
                  errors: list[str] = None,
                  logger: Logger = None) -> None:
 
@@ -45,9 +45,6 @@ class PySob:
         # maps to the entity's PK in its DB table (returned on INSERT operations)
         self.id: int | str | None = None
 
-        # determine whether this instance exists in the database
-        self._is_new: bool = True
-
         # make sure to have an errors list
         if not isinstance(errors, list):
             errors = []
@@ -58,6 +55,7 @@ class PySob:
                       omit_nulls=True,
                       db_engine=db_engine,
                       db_conn=db_conn,
+                      committable=committable,
                       errors=errors)
 
     def insert(self,
@@ -68,12 +66,12 @@ class PySob:
 
         # prepare data for INSERT
         return_col: dict[str, type] | None = None
-        insert_data: dict[str, Any] = self.to_columns(omit_nulls=True)
+        insert_data: dict[str, Any] = self.get()
         cls_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
         is_identity: bool = sob_db_specs[cls_name][2]
         if is_identity:
             # PK is an identity column
-            pk_name: str = sob_col_names[cls_name][0]
+            pk_name: str = sob_db_columns[cls_name][0]
             pk_type: type = sob_db_specs[cls_name][1]
             insert_data.pop(pk_name, None)
             return_col = {pk_name: pk_type}
@@ -93,7 +91,6 @@ class PySob:
                                     errors=errors,
                                     logger=self._logger)
         if not errors:
-            self._is_new = False
             if is_identity:
                 # PK is an identity column
                 self.id = rec[0]
@@ -110,9 +107,9 @@ class PySob:
 
         # prepare data for UPDATE
         cls_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
-        update_data: dict[str, Any] = self.to_columns(omit_nulls=False)
+        pk_name: str = sob_db_columns[cls_name][0]
         tbl_name: str = sob_db_specs[cls_name][0]
-        pk_name = sob_col_names[cls_name][0]
+        update_data: dict[str, Any] = self.get(omit_nulls=False)
         key: int | str = update_data.pop(pk_name)
 
         # make sure to have an errors list
@@ -145,13 +142,13 @@ class PySob:
         # declare the return variable
         result: bool
 
-        if self._is_new:
-            result = self.insert(db_engine=db_engine,
+        if self.id:
+            result = self.update(db_engine=db_engine,
                                  db_conn=db_conn,
                                  committable=committable,
                                  errors=errors)
         else:
-            result = self.update(db_engine=db_engine,
+            result = self.insert(db_engine=db_engine,
                                  db_conn=db_conn,
                                  committable=committable,
                                  errors=errors)
@@ -165,12 +162,12 @@ class PySob:
 
         cls_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
         where_data: dict[str, Any]
+        pk_name: str = sob_db_columns[cls_name][0]
         tbl_name: str = sob_db_specs[cls_name][0]
-        pk_name: str = sob_col_names[cls_name][0]
         if self.id:
             where_data = {pk_name: self.id}
         else:
-            where_data = self.to_columns(omit_nulls=True)
+            where_data = self.get()
             where_data.pop(pk_name, None)
 
         # make sure to have an errors list
@@ -198,38 +195,65 @@ class PySob:
         for key in self.__dict__:
             self.__dict__[key] = None
 
+    def get(self,
+            omit_nulls: bool = True) -> dict[str, Any]:
+
+        # initialize the return variable
+        result: dict[str, Any] = {}
+
+        cls_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
+        if not (omit_nulls and self.id is None):
+            # PK attribute in DB table might have a different name
+            pk_name: str = sob_db_columns[cls_name][0]
+            result[pk_name] = self.id
+        result.update({k: v for k, v in self.__dict__.items()
+                      if k.islower() and not (k.startswith("_") or k == "id" or (omit_nulls and v is None))})
+        return result
+
     def set(self,
             data: dict[str, Any]) -> None:
 
         cls_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
         tbl_name = sob_db_specs[cls_name][0]
         for key, value in data.items():
-            attr: str = (sob_attrs_map.get(cls_name) or {}).get(key) or key
-
-            # use enum names assigned as values in 'data'
             if isinstance(value, Enum) and "use_names" in value.__class__:
-                value = value.name  # noqa: PLW2901
-
-            if attr in self.__dict__:
-                self.__dict__[attr] = value
+                # use enum names assigned as values in 'data'
+                value = value.name
+            if key in self.__dict__:
+                self.__dict__[key] = value
             elif self._logger:
-                self._logger.warning(msg=f"'{attr}'is not an attribute of {tbl_name}")
+                self._logger.warning(msg=f"'{key}' is not an attribute of '{tbl_name}'")
 
-    def is_new(self) -> bool:
+    def get_inputs(self) -> dict[str, Any] | None:
 
-        return self._is_new
+        # initialize the return variable
+        result: dict[str, Any] | None = None
 
-    def is_in_db(self,
-                 db_engine: DbEngine = None,
-                 db_conn: Any = None,
-                 errors: list[str] = None) -> bool | None:
+        # obtain the mapping of input names to attributes
+        cls_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
+        mapping: list[tuple[str, str]] = sob_attrs_input.get(cls_name)
+        if mapping:
+            # obtain the instance's non-null data
+            data: dict[str, Any] = self.get()
+            if data:
+                result = dict_clone(source=data,
+                                    from_to_keys=[(t[1], t[0]) for t in mapping if t[1]])
+        return result
+
+    def is_persisted(self,
+                     db_engine: DbEngine = None,
+                     db_conn: Any = None,
+                     committable: bool = None,
+                     errors: list[str] = None) -> bool | None:
 
         # initialize the return variable
         result: bool = False
 
         cls_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
+        pk_name: str = sob_db_columns[cls_name][0]
         tbl_name: str = sob_db_specs[cls_name][0]
-        pk_name: str = sob_col_names[cls_name][0]
+
+        # build the WHERE clause
         where_data: dict[str, Any] | None = None
         if self.id:
             # use object's ID
@@ -248,37 +272,40 @@ class PySob:
 
         if not where_data:
             # use object's available data
-            where_data = self.to_columns(omit_nulls=True)
+            where_data = self.get()
             where_data.pop(pk_name, None)
 
-        if not where_data:
+        # execute the query
+        if where_data:
             result = db_exists(table=tbl_name,
                                where_data=where_data,
                                engine=db_engine,
                                connection=db_conn,
+                               committable=committable,
                                errors=errors,
                                logger=self._logger)
         return result
 
     def load(self,
-             __references: list[type] = None,
+             __references: type[Sob | list[Sob]] | list[type[Sob | list[Sob]]] = None,
              /,
              omit_nulls: bool = True,
              db_engine: DbEngine = None,
              db_conn: Any = None,
+             committable: bool = None,
              errors: list[str] = None) -> bool:
 
         # initialize the return variable
         result: bool = False
 
         cls_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
+        pk_name: str = sob_db_columns[cls_name][0]
         tbl_name: str = sob_db_specs[cls_name][0]
-        pk_name: str = sob_col_names[cls_name][0]
         where_data: dict[str, Any]
         if self.id:
             where_data = {pk_name: self.id}
         else:
-            where_data = self.to_columns(omit_nulls=omit_nulls)
+            where_data = self.get(omit_nulls=omit_nulls)
             where_data.pop(pk_name, None)
 
         # make sure to have an errors list
@@ -287,12 +314,12 @@ class PySob:
 
         # loading the object from the database might fail
         attrs: list[str] = self.get_columns()
-        recs: list[tuple] = db_select(sel_stmt=f"SELECT {', '.join(attrs)} "
-                                               f"FROM {tbl_name}",
+        recs: list[tuple] = db_select(sel_stmt=f"SELECT {', '.join(attrs)} FROM {tbl_name}",
                                       where_data=where_data,
                                       limit_count=2,
                                       engine=db_engine,
                                       connection=db_conn,
+                                      committable=committable,
                                       errors=errors,
                                       logger=self._logger)
         msg: str | None = None
@@ -318,12 +345,12 @@ class PySob:
                     self.__dict__["id"] = rec[inx]
                 else:
                     self.__dict__[attr] = rec[inx]
-            self._is_new = False
             if __references:
                 PySob.__load_references(__references,
                                         objs=[self],
                                         db_engine=db_engine,
                                         db_conn=db_conn,
+                                        committable=committable,
                                         errors=errors)
             result = not errors
 
@@ -331,49 +358,18 @@ class PySob:
 
     def get_columns(self) -> list[str]:
 
-        # PK attribute in DB table might have a different : list[str] = []
+        # PK attribute in DB table might have a different name
         return [k for k in self.__dict__
                 if k.islower() and not k.startswith("_")]
 
-    def to_columns(self,
-                   omit_nulls: bool) -> dict[str, Any]:
-
-        cls_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
-        pk_name: str = sob_col_names[cls_name][0]
-        # PK attribute in DB table might have a different name
-        result: dict[str, Any] = {pk_name: self.__dict__.get("id")}
-        result.update({k: v for k, v in self.__dict__.items()
-                      if k.islower() and not (k.startswith("_") or k == "id" or (omit_nulls and v is None))})
-        return result
-
-    def to_params(self,
-                  omit_nulls: bool) -> dict[str, Any]:
-
-        return self.data_to_params(data=self.__dict__,
-                                   omit_nulls=omit_nulls)
-
-    def data_to_params(self,
-                       data: dict[str, Any],
-                       omit_nulls: bool) -> dict[str, Any]:
-
-        # initialize the return variable
-        result: dict[str, Any] = {}
-        
-        cls_name: str = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
-        for k, v in data.items():
-            if not omit_nulls or v is not None:
-                attr: str = dict_get_key(source=sob_attrs_map.get(cls_name) or {},
-                                         value=k) or k
-                result[attr] = v
-
-        return result
-
     # noinspection PyUnusedLocal
+    # ruff: noqa: ARG002
     def load_references(self,
-                        __references: type[Sob] | type[list[Sob]] | list[type[Sob] | type[list[Sob]]],
+                        __references: type[Sob | list[Sob]] | list[type[Sob | list[Sob]]],
                         /,
                         db_engine: DbEngine = None,
                         db_conn: Any = None,
+                        committable: bool = None,
                         errors: list[str] = None) -> None:
 
         # must be implemented by subclasses containing references
@@ -385,11 +381,13 @@ class PySob:
             self._logger.error(msg=msg)
 
     # noinspection PyUnusedLocal
+    # ruff: noqa: ARG002
     def invalidate_references(self,
-                              __references: type[Sob] | type[list[Sob]] | list[type[Sob] | type[list[Sob]]],
+                              __references: type[Sob | list[Sob]] | list[type[Sob | list[Sob]]],
                               /,
                               db_engine: DbEngine = None,
                               db_conn: Any = None,
+                              committable: bool = None,
                               errors: list[str] = None) -> None:
 
         # must be implemented by subclasses containing references
@@ -400,11 +398,12 @@ class PySob:
         if self._logger:
             self._logger.error(msg=msg)
 
+    # noinspection PyPep8
     @staticmethod
     def initialize(db_specs: tuple[type[StrEnum] | list[str], int | str] |
-                             tuple[type[StrEnum] | list[str], int, bool],  # noqa
-                   attrs_map: dict[str, str] = None,
+                             tuple[type[StrEnum] | list[str], int, bool],
                    attrs_unique: list[tuple[str]] = None,
+                   attrs_input: list[tuple[str, str]] = None,
                    logger: Logger = None) -> None:
 
         # obtain the invoking class
@@ -420,7 +419,7 @@ class PySob:
             tbl_name: str = attrs.pop(0)
 
             # register the names of DB columnns
-            sob_col_names.update({cls_name: tuple(attrs)})
+            sob_db_columns.update({cls_name: tuple(attrs)})
 
             # register the DB specs (table, PK type, PK entity state)
             if len(db_specs) == 2:
@@ -428,26 +427,29 @@ class PySob:
                 db_specs += (db_specs[1] is int,)
             sob_db_specs.update({cls_name: (tbl_name, db_specs[1], db_specs[2])})
 
-            # register the mapping of input parameters to DB columns
-            if attrs_map:
-                sob_attrs_map.update({cls_name: attrs_map})
-
             # register the sets of unique attributes
             if attrs_unique:
                 sob_attrs_unique.update({cls_name: attrs_unique})
 
+            # register the names used for data input
+            if attrs_input:
+                sob_attrs_input.update({cls_name: attrs_input})
+
             if logger:
                 logger.debug(msg=f"Inicialized access data for class '{cls_name}'")
 
+    # noinspection PyPep8
     @staticmethod
-    def count(table_alias: str = None,
-              table_joins: str = None,
+    def count(alias: str = None,
+              joins: list[tuple[type[Sob] | str, str, str]] |
+                     list[tuple[type[Sob] | str, str, str, Literal["inner", "full", "left", "right"]]] = None,
               count_clause: str = None,
               where_clause: str = None,
               where_vals: tuple = None,
               where_data: dict[str, Any] = None,
               db_engine: DbEngine = None,
               db_conn: Any = None,
+              committable: bool = None,
               errors: list[str] = None,
               logger: Logger = None) -> int | None:
 
@@ -462,19 +464,19 @@ class PySob:
         cls: type[Sob] = PySob.__get_invoking_class(errors=errors,
                                                     logger=logger)
         if not errors:
-            cls_name: str = f"{cls.__module__}.{cls.__qualname__}"
-            tbl_name: str = sob_db_specs[cls_name][0]
-            if table_alias:
-                tbl_name = f"{tbl_name} {table_alias}"
-            if table_joins:
-                tbl_name += f" {table_joins}"
-            result = db_count(table=tbl_name,
+            # build the FROM clause
+            from_clause: str = PySob.__build_from_clause(cls=cls,
+                                                         alias=alias,
+                                                         joins=joins)
+            # retrieve the data
+            result = db_count(table=from_clause,
                               count_clause=count_clause,
                               where_clause=where_clause,
                               where_vals=where_vals,
                               where_data=where_data,
                               engine=db_engine,
                               connection=db_conn,
+                              committable=committable,
                               errors=errors,
                               logger=logger)
         if errors and logger:
@@ -482,9 +484,11 @@ class PySob:
 
         return result
 
+    # noinspection PyPep8
     @staticmethod
-    def exists(table_alias: str = None,
-               table_joins: str = None,
+    def exists(alias: str = None,
+               joins: list[tuple[type[Sob] | str, str, str]] |
+                      list[tuple[type[Sob] | str, str, str, Literal["inner", "full", "left", "right"]]] = None,
                where_clause: str = None,
                where_vals: tuple = None,
                where_data: dict[str, Any] = None,
@@ -492,8 +496,9 @@ class PySob:
                max_count: int = None,
                db_engine: DbEngine = None,
                db_conn: Any = None,
+               committable: bool = None,
                errors: list[str] = None,
-               logger: Logger = None) -> int | None:
+               logger: Logger = None) -> bool | None:
 
         # inicialize the return variable
         result: bool | None = None
@@ -506,13 +511,12 @@ class PySob:
         cls: type[Sob] = PySob.__get_invoking_class(errors=errors,
                                                     logger=logger)
         if not errors:
-            cls_name: str = f"{cls.__module__}.{cls.__qualname__}"
-            tbl_name: str = sob_db_specs[cls_name][0]
-            if table_alias:
-                tbl_name = f"{tbl_name} {table_alias}"
-            if table_joins:
-                tbl_name += f" {table_joins}"
-            result = db_exists(table=tbl_name,
+            # build the FROM clause
+            from_clause: str = PySob.__build_from_clause(cls=cls,
+                                                         alias=alias,
+                                                         joins=joins)
+            # execute the query
+            result = db_exists(table=from_clause,
                                where_clause=where_clause,
                                where_vals=where_vals,
                                where_data=where_data,
@@ -520,6 +524,7 @@ class PySob:
                                max_count=max_count,
                                engine=db_engine,
                                connection=db_conn,
+                               committable=committable,
                                errors=errors,
                                logger=logger)
         if errors and logger:
@@ -527,11 +532,65 @@ class PySob:
 
         return result
 
+    # noinspection PyPep8
     @staticmethod
-    def retrieve(__references: list[type] = None,
+    def get_values(attrs: tuple[str],
+                   alias: str = None,
+                   joins: list[tuple[type[Sob] | str, str, str]] |
+                          list[tuple[type[Sob] | str, str, str, Literal["inner", "full", "left", "right"]]] = None,
+                   where_clause: str = None,
+                   where_vals: tuple = None,
+                   where_data: dict[str, Any] = None,
+                   orderby_clause: str = None,
+                   min_count: int = None,
+                   max_count: int = None,
+                   offset_count: int = None,
+                   limit_count: int = None,
+                   db_engine: DbEngine = None,
+                   db_conn: Any = None,
+                   committable: bool = None,
+                   errors: list[str] = None,
+                   logger: Logger = None) -> list[tuple] | None:
+
+        # inicialize the return variable
+        result: list[tuple] | None = None
+
+        # make sure to have an errors list
+        if not isinstance(errors, list):
+            errors = []
+
+        # obtain the invoking class
+        cls: type[Sob] = PySob.__get_invoking_class(errors=errors,
+                                                    logger=logger)
+        if not errors:
+            # build the FROM clause
+            from_clause: str = PySob.__build_from_clause(cls=cls,
+                                                         alias=alias,
+                                                         joins=joins)
+            # retrieve the data
+            result = db_select(sel_stmt=f"SELECT DISTINCT {', '.join(attrs)} FROM {from_clause}",
+                               where_clause=where_clause,
+                               where_vals=where_vals,
+                               where_data=where_data,
+                               orderby_clause=orderby_clause,
+                               min_count=min_count,
+                               max_count=max_count,
+                               offset_count=offset_count,
+                               limit_count=limit_count,
+                               engine=db_engine,
+                               connection=db_conn,
+                               committable=committable,
+                               errors=errors,
+                               logger=logger)
+        return result
+
+    # noinspection PyPep8
+    @staticmethod
+    def retrieve(__references: type[Sob | list[Sob]] | list[type[Sob | list[Sob]]] = None,
                  /,
-                 table_alias: str = None,
-                 table_joins: str = None,
+                 alias: str = None,
+                 joins: list[tuple[type[Sob] | str, str, str]] |
+                        list[tuple[type[Sob] | str, str, str, Literal["inner", "full", "left", "right"]]] = None,
                  where_clause: str = None,
                  where_vals: tuple = None,
                  where_data: dict[str, Any] = None,
@@ -557,19 +616,21 @@ class PySob:
         cls: type[Sob] = PySob.__get_invoking_class(errors=errors,
                                                     logger=logger)
         if not errors:
-            # build the attributes list
             cls_name: str = f"{cls.__module__}.{cls.__qualname__}"
-            tbl_name: str = sob_db_specs[cls_name][0]
+
+            # build the FROM clause
+            from_clause: str = PySob.__build_from_clause(cls=cls,
+                                                         alias=alias,
+                                                         joins=joins)
+            # build the attributes list
             attrs: list[str] = []
-            for attr in sob_col_names.get(cls_name):
-                if table_alias:
-                    attr = f"{table_alias}.{attr}"
+            for attr in sob_db_columns.get(cls_name):
+                if alias:
+                    attr = f"{alias}.{attr}"
                 attrs.append(attr)
-            if table_alias:
-                tbl_name = f"{tbl_name} {table_alias}"
-            if table_joins:
-                tbl_name += f" {table_joins}"
-            sel_stmt: str = f"SELECT DISTINCT {', '.join(attrs)} FROM {tbl_name}"
+
+            # retrieve the data
+            sel_stmt: str = f"SELECT DISTINCT {', '.join(attrs)} FROM {from_clause}"
             recs: list[tuple[int | str]] = db_select(sel_stmt=sel_stmt,
                                                      where_clause=where_clause,
                                                      where_vals=where_vals,
@@ -589,11 +650,10 @@ class PySob:
                 objs: list[Sob] = []
                 for rec in recs:
                     data: dict[str, Any] = {}
-                    for inx, attr in enumerate(sob_col_names.get(cls_name)):
+                    for inx, attr in enumerate(sob_db_columns.get(cls_name)):
                         data[attr] = rec[inx]
                     sob: type[Sob] = cls()
                     sob.set(data=data)
-                    sob._is_new = False
                     if errors:
                         break
                     objs.append(sob)
@@ -603,6 +663,7 @@ class PySob:
                                             objs=objs,
                                             db_engine=db_engine,
                                             db_conn=db_conn,
+                                            committable=committable,
                                             errors=errors)
                 if not errors:
                     result = objs
@@ -612,42 +673,6 @@ class PySob:
 
         return result
 
-    @staticmethod
-    def store(insert_data: dict[str, Any] = None,
-              return_cols: dict[str, Any] = None,
-              db_engine: DbEngine = None,
-              db_conn: Any = None,
-              committable: bool = None,
-              errors: list[str] = None,
-              logger: Logger = None) -> tuple | int | None:
-
-        # initialize the return variable
-        result: tuple | int | None = None
-
-        # make sure to have an errors list
-        if not isinstance(errors, list):
-            errors = []
-
-        # obtain the invoking class
-        cls: type[Sob] = PySob.__get_invoking_class(errors=errors,
-                                                    logger=logger)
-        # delete specified rows
-        if not errors:
-            cls_name: str = f"{cls.__module__}.{cls.__qualname__}"
-            tbl_name: str = sob_db_specs[cls_name][0]
-            result = db_insert(insert_stmt=f"INSERT INTO {tbl_name}",
-                               insert_data=insert_data,
-                               return_cols=return_cols,
-                               engine=db_engine,
-                               connection=db_conn,
-                               committable=committable,
-                               errors=errors,
-                               logger=logger)
-        if errors and logger:
-            logger.error(msg="; ".join(errors))
-
-        return result
-
     @staticmethod
     def erase(where_clause: str = None,
               where_vals: tuple = None,
@@ -690,6 +715,42 @@ class PySob:
 
         return result
 
+    @staticmethod
+    def store(insert_data: dict[str, Any] = None,
+              return_cols: dict[str, Any] = None,
+              db_engine: DbEngine = None,
+              db_conn: Any = None,
+              committable: bool = None,
+              errors: list[str] = None,
+              logger: Logger = None) -> tuple | int | None:
+
+        # initialize the return variable
+        result: tuple | int | None = None
+
+        # make sure to have an errors list
+        if not isinstance(errors, list):
+            errors = []
+
+        # obtain the invoking class
+        cls: type[Sob] = PySob.__get_invoking_class(errors=errors,
+                                                    logger=logger)
+        # delete specified rows
+        if not errors:
+            cls_name: str = f"{cls.__module__}.{cls.__qualname__}"
+            tbl_name: str = sob_db_specs[cls_name][0]
+            result = db_insert(insert_stmt=f"INSERT INTO {tbl_name}",
+                               insert_data=insert_data,
+                               return_cols=return_cols,
+                               engine=db_engine,
+                               connection=db_conn,
+                               committable=committable,
+                               errors=errors,
+                               logger=logger)
+        if errors and logger:
+            logger.error(msg="; ".join(errors))
+
+        return result
+
     @staticmethod
     def __get_invoking_class(errors: list[str] | None,
                              logger: Logger | None) -> type[Sob] | None:
@@ -741,7 +802,7 @@ class PySob:
 
         if not result:
             msg: str = (f"Unable to obtain class '{classname}', "
-                        f"filepath '{filepath}', "f"from invoking function '{invoking_function}'")
+                        f"filepath '{filepath}', from invoking function '{invoking_function}'")
             if logger:
                 logger.error(msg=f"{msg} - invocation frame {caller_frame}")
             if isinstance(errors, list):
@@ -750,22 +811,23 @@ class PySob:
         return result
 
     @staticmethod
-    def __load_references(__references: list[type],
+    def __load_references(__references:  type[Sob | list[Sob]] | list[type[Sob | list[Sob]]],
                           /,
                           objs: list[Sob],
-                          db_engine: DbEngine = None,
-                          db_conn: Any = None,
-                          errors: list[str] = None) -> None:
+                          db_engine: DbEngine | None,
+                          db_conn: Any | None,
+                          committable: bool | None,
+                          errors: list[str]) -> None:
 
         if SOB_MAX_THREADS > 1 and \
-                (len(objs) > 1 or len(__references) > 1):
+                (len(objs) > 1 or (isinstance(__references, list) and len(__references) > 1)):
             task_futures: list[Future] = []
             with ThreadPoolExecutor(max_workers=SOB_MAX_THREADS) as executor:
                 for obj in objs:
-                    for __reference in __references:
+                    for reference in __references if isinstance(__references, list) else [__references]:
                         # must not multiplex 'db_conn'
                         future: Future = executor.submit(obj.load_references,
-                                                         [__reference],
+                                                         reference,
                                                          db_engine=db_engine,
                                                          errors=errors)
                         if errors:
@@ -782,6 +844,35 @@ class PySob:
                 obj.load_references(__references,
                                     db_engine=db_engine,
                                     db_conn=db_conn,
+                                    committable=committable,
                                     errors=errors)
                 if errors:
                     break
+
+    # noinspection PyPep8
+    @staticmethod
+    def __build_from_clause(alias: str,
+                            cls: type[Sob],
+                            joins: list[tuple[type[Sob] | str, str, str]] |
+                                   list[tuple[type[Sob] | str, str, str,
+                                              Literal["inner", "full", "left", "right"]]] | None) -> str:
+
+        # obtain the the fully-qualified name of the class type
+        cls_name: str = f"{cls.__module__}.{cls.__qualname__}"
+
+        # establish the main query table
+        result: str = sob_db_specs[cls_name][0]
+        if alias:
+            result += " AS " + alias
+
+        # build the joins
+        for join in joins or []:
+            if isinstance(join[0], str):
+                target: str = join[0]
+            else:
+                name: str = f"{join[0].__module__}.{join[0].__qualname__}"
+                target: str = sob_db_specs[name][0]
+            mode: str = join[3].upper() if len(join) > 3 else "INNER"
+            result += f" {mode} JOIN {target} AS {join[1]} ON {join[2]}"
+
+        return result
